/* This class extends Capture, a processing class specialised in grabbing camera
 * Methods added allow the background substraction and blob detection to be performed
 * at each "read()" call, synchronised with "captureEvent"
 * Display can be slower then the data retrieval.
 * To display the current image, use "image( myCapture.getImg(), 0, 0 )", not "image( myCapture, 0, 0 )".
 * The processing done during the "read" makes the normal display a bit laggy.
 * The x latest frames of blobs are available in frames array.
 * To get it in the main class, use something like: "ArrayList<float[]> frames = myCapture.cloneFrames();"; 
 * this will return a COPY of internal frames.
 * DO NOT TRY TO DRAW BEFORE THE BlobedCapture IS UNLOCKED -> it will cause a concurrent access exception!
 * To do so, use
 * while( myCapture.isLocked() ) {
 * 	try {
 *		Thread.sleep( 1 );
 *	} catch (InterruptedException e) {
 *		e.printStackTrace();
 *	}
 * }
 * During processing, BlobedCapture blocks the draw of the main class.
 * No concurrent access exception.
 * Good luck!
 * */

import java.util.ArrayList;

import processing.core.PApplet;
import processing.core.PImage;
import processing.video.Capture;
import processing.blobDetection.*;

public class BlobedCapture extends Capture {

	public static final float CONNECTION_RADIUS_DEFAULT = 0.35f;
	public static final float BLOB_MINIMUM_SIZE_DEFAULT = 0.02f;
	public static final float BLOB_MAXIMUM_SIZE_DEFAULT = 0.4f;
	public static final float BLOB_THRESHOLD_DEFAULT = 0.23f;
	public static final int BACKGROUND_TOLERANCE_DEFAULT = 50;
	public static final int FRAMES_STACK_NUM_DEFAULT = 20;
	
	private float connectionradius;
	private float blobminsize;
	private float blobmaxsize;
	private float blobthreshold;
	private int backgroundtolerance;

	private PImage diff;
	private PImage img;
	private BlobDetection blobdetector;
	private int framestacknum;
	
	/* Contains information about the 20 (framestacknum) last packs of blobs
	 * the length is always a mult of 5
	 * position:
	 * 0: center x
	 * 1: center y
	 * 2: width
	 * 3: height
	 * 4: position of the closest blob in the previous frame
	 * For first frame, this index is always -1
	 * */
	private ArrayList< float[] > frames;
	
	// avoid concurrent modifications
	private boolean locked;

	private int processtime;
	
	public BlobedCapture(PApplet parent, int requestWidth, int requestHeight, String cameraName, int frameRate) {
		
		super(parent, requestWidth, requestHeight, cameraName, frameRate);
		
		connectionradius = CONNECTION_RADIUS_DEFAULT;
		blobminsize = BLOB_MINIMUM_SIZE_DEFAULT;
		blobmaxsize = BLOB_MAXIMUM_SIZE_DEFAULT;
		blobthreshold = BLOB_THRESHOLD_DEFAULT;
		backgroundtolerance = BACKGROUND_TOLERANCE_DEFAULT;
		
		diff = new PImage( this.width, this.height );
		img = new PImage( this.width, this.height );
		
		blobdetector = new BlobDetection( this.width, this.height );
		blobdetector.setPosDiscrimination( true );
		blobdetector.setThreshold( blobthreshold );
		
		framestacknum = FRAMES_STACK_NUM_DEFAULT;
		frames = new ArrayList< float[] >();
		
		processtime = 0;
		
	}
	
	public void configure( 
			int backgroundtolerance, 
			int framestacknum, 
			float blobminsize, 
			float blobmaxsize, 
			float blobthreshold, 
			float connectionradius ) {
		this.backgroundtolerance = backgroundtolerance;
		this.framestacknum = framestacknum;
		this.blobminsize = blobminsize;
		this.blobmaxsize = blobmaxsize;
		if ( this.blobthreshold != blobthreshold ) {
			blobdetector.setThreshold( blobthreshold );
			this.blobthreshold = blobthreshold;
		}
		this.connectionradius = connectionradius;
	}
	
	public synchronized void read() {
		
		int ms1 = parent.millis();
		
		locked = true;
		
		super.read();
		processDiff();
		fastblur( diff, 3 );
		blobdetector.computeBlobs( diff.pixels );
		// process frames
		processFrame();
		
		locked = false;
		
		int ms2 = parent.millis();
		processtime = ms2 - ms1;
		
	}
	
	private int connect( Blob b ) {
		if ( frames.size() == 0 ) return -1;
		float[] prevframe = frames.get( frames.size() - 1 );
		float closest = 0;
		int sel = -1;
		for ( int i = 0; i < prevframe.length; i += 5 ) {
			float hw = prevframe[ i + 2 ] * 0.5f;
			float hh = prevframe[ i + 3 ] * 0.5f;
			// 4 points comparaison
			float d = (
					PApplet.dist( prevframe[ i ] - hw, prevframe[ i + 1 ] - hh, b.x,  b.y ) +
					PApplet.dist( prevframe[ i ] + hw, prevframe[ i + 1 ] - hh, b.x + b.w,  b.y ) +
					PApplet.dist( prevframe[ i ] + hw, prevframe[ i + 1 ] + hh, b.x + b.w,  b.y + b.h ) +
					PApplet.dist( prevframe[ i ] - hw, prevframe[ i + 1 ] + hh, b.x,  b.y + b.h )
					) * 0.25f;
			if ( 
				( closest == 0 && d <= connectionradius ) ||
				( closest > d )
				) {
				sel = i;
				closest = d;
			}
		}
		return sel;
	}
	
	private void processFrame() {
		
		Blob b;
		ArrayList< Blob > validblobs = new ArrayList< Blob >();
		for ( int n = 0; n < blobdetector.getBlobNb(); ++n ) {
			b = blobdetector.getBlob(n);
			if ( b != null && 
					b.w >= blobminsize && b.h >= blobminsize && 
					b.w <= blobmaxsize && b.h <= blobmaxsize		
				) validblobs.add( b );
		}
		float[] frame = new float[ validblobs.size() * 5 ];
		int i = 0;
		for ( Blob bl : validblobs ) {
			// center
			frame[ i ] = bl.x + bl.w * 0.5f;
			frame[ i + 1 ] = bl.y + bl.h * 0.5f;
			// size
			frame[ i + 2 ] = bl.w;
			frame[ i + 3 ] = bl.h;
			// parenting
			frame[ i + 4 ] = connect( bl );
			i += 5;
		}
		frames.add( frame );
		while( frames.size() > framestacknum ) frames.remove( 0 );
		
	}
	
	private void processDiff() {
		
		int pxnum = this.width * this.height;
		this.loadPixels();
		img.loadPixels();
		diff.loadPixels();
		int white = parent.color( 255,255,255 );
		int black = parent.color( 0,0,0 );
		for ( int i = 0; i < pxnum; ++i ) {
			int ic = this.pixels[ i ];
			int pc = img.pixels[ i ];
			if (
				parent.abs( ( ( ic >> 16) & 0xFF ) - ( ( pc >> 16) & 0xFF ) ) > backgroundtolerance ||
				parent.abs( ( ( ic >> 8) & 0xFF ) - ( ( pc >> 8) & 0xFF ) ) > backgroundtolerance ||
				parent.abs( ( ic & 0xFF ) - ( pc & 0xFF ) ) > backgroundtolerance
			) {
				diff.pixels[ i ] = white;
			} else {
				diff.pixels[ i ] = black;
			}
			// backuping at the same time
			img.pixels[ i ] = ic;
		}
		img.updatePixels();
		diff.updatePixels();
	
	}
	
	private void fastblur(PImage img, int radius) {
		
		if (radius < 1) {
			return;
		}
		
		int w = img.width;
		int h = img.height;
		int wm = w - 1;
		int hm = h - 1;
		int wh = w * h;
		int div = radius + radius + 1;
		int r[] = new int[wh];
		int g[] = new int[wh];
		int b[] = new int[wh];
		int rsum, gsum, bsum, x, y, i, p, p1, p2, yp, yi, yw;
		int vmin[] = new int[PApplet.max(w, h)];
		int vmax[] = new int[PApplet.max(w, h)];
		int[] pix = img.pixels;
		int dv[] = new int[256 * div];
		for (i = 0; i < 256 * div; i++) {
			dv[i] = (i / div);
		}
		yw = yi = 0;

		for (y = 0; y < h; y++) {
			rsum = gsum = bsum = 0;
			for (i = -radius; i <= radius; i++) {
				p = pix[yi + PApplet.min(wm, PApplet.max(i, 0))];
				rsum += (p & 0xff0000) >> 16;
				gsum += (p & 0x00ff00) >> 8;
				bsum += p & 0x0000ff;
			}
			for (x = 0; x < w; x++) {
				r[yi] = dv[rsum];
				g[yi] = dv[gsum];
				b[yi] = dv[bsum];
				if (y == 0) {
					vmin[x] = PApplet.min(x + radius + 1, wm);
					vmax[x] = PApplet.max(x - radius, 0);
				}
				p1 = pix[yw + vmin[x]];
				p2 = pix[yw + vmax[x]];
				rsum += ((p1 & 0xff0000) - (p2 & 0xff0000)) >> 16;
				gsum += ((p1 & 0x00ff00) - (p2 & 0x00ff00)) >> 8;
				bsum += (p1 & 0x0000ff) - (p2 & 0x0000ff);
				yi++;
			}
			yw += w;
		}

		for (x = 0; x < w; x++) {
			rsum = gsum = bsum = 0;
			yp = -radius * w;
			for (i = -radius; i <= radius; i++) {
				yi = PApplet.max(0, yp) + x;
				rsum += r[yi];
				gsum += g[yi];
				bsum += b[yi];
				yp += w;
			}
			yi = x;
			for (y = 0; y < h; y++) {
				pix[yi] = 0xff000000 | (dv[rsum] << 16) | (dv[gsum] << 8) | dv[bsum];
				if (x == 0) {
					vmin[y] = PApplet.min(y + radius + 1, hm) * w;
					vmax[y] = PApplet.max(y - radius, 0) * w;
				}
				p1 = x + vmin[y];
				p2 = x + vmax[y];
				rsum += r[p1] - r[p2];
				gsum += g[p1] - g[p2];
				bsum += b[p1] - b[p2];
				yi += w;
			}
		}
		
	}


	public ArrayList< float[] > cloneFrames() {
		return (ArrayList<float[]>) frames.clone();
	}
	
	public PImage getDiff() {
		return diff;
	}

	public PImage getImg() {
		return img;
	}

	public int getProcesstime() {
		return processtime;
	}

	public int getFramestacknum() {
		return framestacknum;
	}

	public boolean isLocked() {
		return locked;
	}
	
	
}
